<?php

// DBTable.class.php - PHP-OF database table class
//
// $Id: DBTable.class.php 1870 2009-08-02 20:25:52Z tim $
//
// (c)2001-2008 Tim Jackson (tim@timj.co.uk)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

require_once('DB-phpof.php');

define('DBTABLE_COLUMNS_BINARY',1);
define('DBTABLE_COLUMNS_NONBINARY',2);
define('DBTABLE_COLUMNS_ALL',3);

define('DBTABLE_COLTYPE_STRING',1);	// ~Equivalent to ADODB 'C' type
define('DBTABLE_COLTYPE_BLOB',2);	// ~Equivalent to ADODB 'B' type
define('DBTABLE_COLTYPE_INT',3);	// ~Equivalent to ADODB 'I' type
define('DBTABLE_COLTYPE_DATE',4);	// ~Equivalent to ADODB 'D' type
define('DBTABLE_COLTYPE_BOOL',5);	// ~Equivalent to ADODB 'L' type
define('DBTABLE_COLTYPE_ENUM',6);
define('DBTABLE_COLTYPE_SET',7);
define('DBTABLE_COLTYPE_TIME',8);
define('DBTABLE_COLTYPE_DATETIME',9);
define('DBTABLE_COLTYPE_TEXT',10);	// ~Equivalent to ADODB 'X' type
define('DBTABLE_COLTYPE_UNKNOWN',11);
define('DBTABLE_COLTYPE_REAL',11);	// ~Equivalent to ADODB 'N' type

define('DBTABLE_COLKEY_FOREIGN','!foreign!');

define('DBTABLE_USEDN',-1);
define('DBTABLE_USEKEY',-2);

// Used in getArray()
define('DBTABLE_COLNAME',1);
define('DBTABLE_EXPRESSION',2);

// Legacy scripts (<= DBTable v1.0.21) may rely on being able to pass in the empty string??
define('DBTABLE_USE_DEFAULT_ORDER','');

/**
 * This class is used for handling database tables. Upon instantiation,
 * information about the columns and primary key within table_name will be
 * automatically fetched from the database and stored internally within the
 * object.
 * 
 * The DBTable class is ideally suited as a parent for creating derived classes
 * from. For example, if you were handling a database of hotels, you could 
 * create a derived class "Hotels" which had functions such as 
 * Hotels::findAvailableRoom() and so on. The key point is that DBTable provides
 * you with functions to ease the core, repetitive database interfacing such as 
 * inserting, updating and deleting rows.
 *
 */
class DBTable {

	var $_name;									// Table name in database
	var $_fields;								// Array containing all fields except $_key and any binary types
	var $_fieldlist;							// String containing comma-separated list of fields except $_key and any binary types
	var $_fieldtype;							// Array containing the fieldtypes except $_key
	var $_binaryfields;						// Array containing a list of binary fields
	var $_binaryfieldlist;					// String containing comma-separated list of binary fields
	var $allownull;		// Array containing list of fields and whether they can be null or not (NEW)
	var $_key;									// Primary key column for this table
	var $_key_auto_increment = false;	// Does the primary key column auto inc?
	var $_db;
	var $_tablecache = false;				// Table cache DBM resource
	var $_SortOrder = '';	// The column containing an integer value which represents the (arbitrary) sort order of the rows
	var $_SortDefaultPosition = null;	// whether to add new rows to top or bottom of table. Set to DBROW_SORT_BOTTOM or DBROW_SORT_TOP. Default is null, which is the equivalent of DBROW_SORT_BOTTOM
	var $SupportsTransactions = false;
	var $RowObjectName = '';	// The object name of DBRow objects which represent rows in this table
	var $DefaultOrder = '';
	// TODO: implement use of DefaultOrder throughout including in DB_result possibly?
	
	// An (optional) associative array containing descriptions of columns
	var $ColumnDescription = false;
	
	// Contains an array of defined keys
	var $ForeignKeys;
	
	// What method is used to discover foreign keys
	var $_ForeignKeyDiscoveryType;
	
	// Cache of foreign keys
	var $_ForeignKeyCache;
	
	// Audit logging enabled
	var $AuditLogging;
	
	// Audit columns to ignore
	var $AuditColumnsIgnore = array();
	
	/**
	 * Constructor
	 * 
	 * @param mixed $db  A PHPOF DB_xxx object
	 * @param string $name  Name of table
	 * @param string $row_name  Name of row objects which are normally used with this table.
	 * @return DBTable
	 */
	function DBTable (&$db, $name, $row_name='')
	{
		global $PHPOF_FOREIGN_KEY_DISCOVERY_TYPE, $PHPOF_AUDIT_LOGGING;
		
		if (!is_object($db)) trigger_error('DBTable constructor called without a valid DB object',E_USER_ERROR);
		if (!$name) trigger_error('DBTable constructor called with an empty table name',E_USER_ERROR);
		$this->_db = &$db;
		$this->_name = $name;
		$this->_getFields();
		
		// Row object name
		if (!empty($row_name)) {
			$this->RowObjectName = $row_name;
		} else if (empty($this->RowObjectName)) {
			$this->RowObjectName = substr($name,0,-1);
			// TODO: more intelligent handling for '...ies'
		}
		
		// Foreign key discovery
		if (!empty($PHPOF_FOREIGN_KEY_DISCOVERY_TYPE)) {
			$this->_ForeignKeyDiscoveryType = $PHPOF_FOREIGN_KEY_DISCOVERY_TYPE;
		} else {
			$this->_ForeignKeyDiscoveryType = 'colnames';
		}
		
		// Audit logging
		if (!isset($this->AuditLogging)) {
			$this->AuditLogging = $PHPOF_AUDIT_LOGGING;
		}
	}
	
	function createRowObject($KeyValue=DBROW_KEY_NONE)
	{
		$RowObject = $this->RowObjectName;
		return new $RowObject($this, $KeyValue);
	}
	
	function _openTableCache ($mode = 'r')
	{
		global $PHPOF_TABLE_CACHING,$PHPOF_TABLE_CACHING_TYPE;
		if ($PHPOF_TABLE_CACHING && !$this->_tablecache) {
			$this->_tablecache = @dba_open($PHPOF_TABLE_CACHING,$mode,$PHPOF_TABLE_CACHING_TYPE);
		}
	}

	function _closeTableCache ()
	{
		if ($this->_tablecache) {
			@dba_close($this->_tablecache);
			unset($this->_tablecache);
		}
	}

	// Publically-accesible function to get a list of fields in the database
	function getFields($Type=DBTABLE_COLUMNS_NONBINARY)
	{
		switch ($Type) {
		case DBTABLE_COLUMNS_NONBINARY:
			return $this->_fields;
			break;
		case DBTABLE_COLUMNS_BINARY:
			return $this->_binaryfields;
		case DBTABLE_COLUMNS_ALL:
			if (count($this->_fieldtype) > 0) {
				foreach ($this->_fieldtype as $FieldName => $Type) {
					$CompleteFieldlist[] = $FieldName;
				}
				return $CompleteFieldlist;
			}
		}
	}
	
	// Returns an array (PHPOF column type, Column extra info, Foreign key table id (or false))
	function getColumnInfo($ColumnName)
	{
		if (empty($ColumnName)) trigger_error('DBTable::getColumnName() called with empty column name',E_USER_ERROR);
		$Column = new DBColumn($ColumnName);
		_DBTable_getColumnInfo($this, $Column);
		
		// Work out foreign keys
		$this->getForeignKeys();	// Will create a cache
		if (isset($this->_ForeignKeyCache[$Column->Name]) && !empty($this->_ForeignKeyCache[$Column->Name])) {
			$Column->ForeignKey = $this->_ForeignKeyCache[$Column->Name];
		}
		return $Column;
	}
	
	function getKeyField()
	{
		if (!empty($this->_key)) {
			return $this->_key;
		} else {
			return false;
		}
	}

	function _getFields ()
	{
		global $PHPOF_TABLE_CACHING,$PHPOF_DEBUG_MODE;
		
		if ($PHPOF_TABLE_CACHING) {
			$this->_openTableCache();
			if ($this->_tablecache) {
				$DBMKey = $this->_db->_dbtype.'://'.$this->_db->_dbhost.'/'.$this->_db->_dbname.'.'.$this->_name;

				if (dba_exists($DBMKey,$this->_tablecache)) {
					$Value = dba_fetch($DBMKey,$this->_tablecache);
					if (!$Value) trigger_error('Null value returned for table cache result in DBTable::_getFields()',E_USER_ERROR);
					if ($PHPOF_DEBUG_MODE) phpof_debug("DBTable::_getFields(): found cached result for table $this->_name");
					$Parts = explode('|',$Value);
					$this->_fieldlist = $Parts[0];
					$this->_binaryfieldlist = $Parts[1];
					$this->_key = $Parts[2];
					$this->_key_auto_increment = $Parts[3];
					$SubParts = explode(',',$Parts[0]);	// Get field list
					if (count($SubParts) > 0) {
						foreach ($SubParts as $Part) {
							if (!empty($Part)) $this->_fields[] = str_replace('"','',$Part);
						}
					}
					$SubParts = explode(',',$Parts[1]);	// Get binary field list
					if (count($SubParts) > 0) {
						foreach ($SubParts as $Part) {
							if (!empty($Part)) $this->_binaryfields[] = str_replace('"','',$Part);
						}
					}
					return;
				}
			}
		}
		_DBTable_GetFields($this);	
		
		// Write to table cache
		if ($PHPOF_TABLE_CACHING && !$Value) {
			$Value = "$this->_fieldlist|$this->_binaryfieldlist|$this->_key|$this->_key_auto_increment";
			$this->_closeTableCache();
			$this->_openTableCache('c');
			if ($this->_tablecache) {
				if ($PHPOF_DEBUG_MODE) phpof_debug("DBTable::_getFields(): Writing table cache for $this->_name");
				dba_insert($DBMKey,$Value,$this->_tablecache);
				@dba_optimize($this->_tablecache);
				$this->_closeTableCache();
			}
		}
	}
		 
	// ------------------------------------------------------
	// Array generation functions

	/**
	 * Retrieve all rows from the database as an associative array
	 * 
	 * This function can be very useful for creating arrays ready to be turned 
	 * into an HTML SELECT using XArray::CreateSelect.
	 * 
	 * Example:
	 * Assume we have an SQL table like so:
	 * +------+----------------+
	 * | code | product_name   |
	 * +------+----------------+
	 * | FB1  | Foo Bar        |
	 * | FB2  | Deluxe Foo Bar |
	 * +------+----------------+
	 * 
	 * $all_products = $mytable->GetArray('code','product_name');
	 * 
	 * $all_products now contains: 
	 * array ( 
	 *    "FB1" => "Foo Bar",
	 *    "FB2" => "Deluxe Foo Bar"
	 * )
	 *
	 * @param string $key  Column in the result set to use as key in associative array
	 * @param string $value  Column in the result set to populate the values in the associative array, or an expression (if $value_type is set to DBTABLE_EXPRESSION)
	 * @param string $orderby  How to order the results. SQL-alike ORDER BY clause.
	 * @param string $value_type  What $value is. DBTABLE_COLNAME means it is a column name; DBTABLE_EXPRESSION means it is an expression.
	 * @return array  The result set in an array
	 * @see DBTable::getArray_paged()
	 * @see DB_result::getArray()
	 */
	function getArray ($key=DBTABLE_USEKEY, $value=DBTABLE_USEDN, $orderby=DBTABLE_USE_DEFAULT_ORDER, $value_type=DBTABLE_COLNAME)
	{
		if ($key == DBTABLE_USEKEY) $key = $this->_key;
		
		// If 'value' is set to DBTABLE_USEDN, then each row is fetched and in turn the Descriptive Name
		// is evaluated according to the DBRow::DescriptiveName property
		if ($value == DBTABLE_USEDN || $value_type == DBTABLE_EXPRESSION) {
			$Row = new $this->RowObjectName($this);
			
			// If $value_type is set to DBTABLE_EXPRESSION then $value is assumed to be the DescriptiveNameExpression
			if ($value_type == DBTABLE_EXPRESSION) {
				$Row->DescriptiveNameExpression = $value;
			}
			
			if (empty($Row->DescriptiveNameExpression)) trigger_error("DBTable::getArray() called with request to use DBRow DescriptiveName (DN) as result value, but DBRow object ($this->RowObjectName) does not have a DescriptiveNameExpression defined");
			$AllRows = $this->getAll($orderby);
			if ($AllRows->numRows() > 0) {
				while ($AllRows->fetchIntoObject($Row)) {
					$ReturnArray[$Row->{$this->_key}] = $Row->DescriptiveName;
				}
				return $ReturnArray;
			}
		} else {
			if ($orderby == DBTABLE_USE_DEFAULT_ORDER) {
				if ($value_type == DBTABLE_COLNAME) {
					$orderby = $this->_getOrderPreference($value);
				} else {
					$orderby = $this->_getOrderPreference($key);
				}
			}
			$res = $this->_db->query("SELECT $key, $value FROM $this->_name ORDER BY $orderby");
			return $res->getArray($key, $value);
		}
	}
	
	/**
	 * As DBTable::getArray() but with paged results
	 *
	 * @param string $key  Column in the result set to use as key in associative array
	 * @param string $value  Column in the result set to populate the values in the associative array, or an expression (if $value_type is set to DBTABLE_EXPRESSION)
	 * @param int $results_per_page   Number of results per page
	 * @param int $page  Page number (starting from 1) of current page of results
	 * @param string $orderby  How to order the results. SQL-alike ORDER BY clause.
	 * @return array  The result set in an array
	 * @see DBTable::getArray()
	 */
	function getArray_Paged ($key, $value, $results_per_page, $page, $orderby=DBTABLE_USE_DEFAULT_ORDER)
	{
		if ($orderby == DBTABLE_USE_DEFAULT_ORDER) $orderby = $this->_getOrderPreference($value);
		$res = $this->_db->query_paged("SELECT $key, $value FROM $this->_name ORDER BY $orderby", $results_per_page, $page);
		return $res->getArray($key, $value);
	}
	
	// Find the preferred ordering for this column (i.e. ORDER BY)
	// The preference is found by taking the following possible sources 
	// of ordering information in turn until a non-empty one is found
	// - the DefaultOrder column
	// - the _SortOrder column (ascending)
	// - the supplied fallback (if any)
	// - the integer 1
	function _getOrderPreference($Fallback=null)
	{
		if (empty($this->DefaultOrder)) {
			if (!empty($this->_SortOrder)) {
				return $this->_SortOrder.' ASC';
			} else if (!empty($Fallback)) {
				return $Fallback;
			} else {
				return '1';
			}
		}
		return $this->DefaultOrder;
	}

	
	// ------------------------------------------------------
	// Search functions
	
	/**
	 * Return all rows from the table
	 *
	 * @param string $orderby  SQL-type "ORDER BY" clause. If supplied, the results will be ordered as specified.
	 * @return DB_result results
	 * @see DBTable::getAll_paged()
	 */
	function getAll ($orderby=DBTABLE_USE_DEFAULT_ORDER)
	{
		return $this->_getAll ($orderby);
	}
	
	/**
	 * As DBTable::getAll() but with paging.
	 *
	 * @param int $results_per_page   Number of results per page
	 * @param int $page  Page number (starting from 1) of current page of results
	 * @param string $orderby  SQL-type "ORDER BY" clause. If supplied, the results will be ordered as specified.
	 * @return DB_result results
	 * @see DBTable::getAll()
	 */
	function getAll_paged ($results_per_page, $page, $orderby=DBTABLE_USE_DEFAULT_ORDER)
	{
		if ($page == 0) $page = 1;
		return $this->_getAll ($orderby, $results_per_page, $page);
	}
	
	function _getAll ($orderby = DBTABLE_USE_DEFAULT_ORDER, $results_per_page = 0, $page = 0)
	{
		if ($orderby == DBTABLE_USE_DEFAULT_ORDER) {
			$orderby = $this->_getOrderPreference();
		}
		if (empty($this->_fieldlist)) trigger_error('DBTable::getAll* called with empty fieldlist in table class',E_USER_ERROR);
		$sql = "SELECT $this->_fieldlist";
		if ($this->_key) $sql .= ",$this->_key";
		$sql .= ' FROM '.$this->_name;
		if ($orderby) $sql .= " ORDER BY $orderby";
		if ($results_per_page > 0) {
			return $this->_db->query_paged($sql,$results_per_page,$page);
		} else {
			return $this->_db->query($sql);
		}
	}
	
	/**
	 * Do a basic table search, looking for $field == $value
	 * 
	 * BasicSearch does an exact match search for a given value in a specified
	 * column.
	 * 
	 * @param string $field  Column to search
	 * @param string $value  Value to search for
	 * @param string $orderby  SQL-type "ORDER BY" clause. If supplied, the results will be ordered as specified.
	 * @return DB_result  Result set
	 */
	function basicSearch ($field, $value, $orderby = DBTABLE_USE_DEFAULT_ORDER)
	{
		return $this->_search('=', $field, $value, $orderby);
	}
	
	/**
	 * As DBTable::BasicSearch but with paged results
	 *
	 * @param string $field  Column to search
	 * @param string $value  Value to search for
	 * @param int $results_per_page   Number of results per page
	 * @param int $page  Page number (starting from 1) of current page of results
	 * @param string $orderby  SQL-type "ORDER BY" clause. If supplied, the results will be ordered as specified.
	 * @return DB_result  Result set
	 * @see DBTable::basicSearch()
	 */
	function basicSearch_Paged ($field, $value, $results_per_page, $page, $orderby = DBTABLE_USE_DEFAULT_ORDER)
	{
		return $this->_search('=', $field, $value, $orderby, $results_per_page, $page);
	}
	
	/**
	 * Do a basic table search, looking for $field != $value
	 * 
	 * This method searches for a cell not matching search_for in 
	 * column search_column. This function is otherwise identical to 
	 * DBTable::basicSearch().
	 *
	 * @param string $field  Column to search
	 * @param string $value  Value to search for
	 * @param string $orderby  SQL-type "ORDER BY" clause. If supplied, the results will be ordered as specified.
	 * @return DB_result  Result set
	 * @see DBTable::basicSearch()
	 */
	function basicSearchNot ($field, $value, $orderby = DBTABLE_USE_DEFAULT_ORDER)
	{
		return $this->_search('!=', $field, $value, $orderby);
	}
	
	/**
	 * As DBTable::basicSearchNot but with paged results
	 *
	 * @param string $field  Column to search
	 * @param string $value  Value to search for
	 * @param int $results_per_page   Number of results per page
	 * @param int $page  Page number (starting from 1) of current page of results
	 * @param string $orderby  SQL-type "ORDER BY" clause. If supplied, the results will be ordered as specified.
	 * @return DB_result  Result set
	 * @see DBTable::basicSearchNot()
	 */
	function basicSearchNot_Paged ($field, $value, $results_per_page, $page, $orderby = DBTABLE_USE_DEFAULT_ORDER)
	{
		return $this->_search('!=', $field, $value, $orderby, $results_per_page, $page);
	}
	
	/**
	 * Similar to DBTable::basicSearch() but does a LIKE instead of an =
	 *
	 * This method returns a database result object when searching for a cell 
	 * LIKE search_for in column search_column. No percentage signs are 
	 * inserted, thereby giving maximum flexibility. Therefore, search_for 
	 * would typically contain something like "%something to search for%".
	 * 
	 * This function is otherwise identical to DBTable::basicSearch().
	 *
	 * @param string $field  Column to search
	 * @param string $value  Value to search for
	 * @param string $orderby  SQL-type "ORDER BY" clause. If supplied, the results will be ordered as specified.
	 * @return DB_result  Result set
	 * @see DBTable::basicSearch()
	 * @see DBTable::basicSearchLikePaged()
	 */
	function basicSearchLike ($field, $value, $orderby = DBTABLE_USE_DEFAULT_ORDER) {
		return $this->_search('LIKE', $field, $value, $orderby);
	}
	
	/**
	 * As DBTable::basicSearchLike but with paged results
	 *
	 * @param string $field  Column to search
	 * @param string $value  Value to search for
	 * @param int $results_per_page   Number of results per page
	 * @param int $page  Page number (starting from 1) of current page of results
	 * @param string $orderby  SQL-type "ORDER BY" clause. If supplied, the results will be ordered as specified.
	 * @return DB_result  Result set
	 * @see DBTable::basicSearchLike()
	 */
	function basicSearchLike_Paged ($field, $value, $results_per_page, $page, $orderby = DBTABLE_USE_DEFAULT_ORDER) {
		return $this->_search('LIKE', $field, $value, $orderby, $results_per_page, $page);
	}
	
	function _search ($clause, $field, $value, $orderby = DBTABLE_USE_DEFAULT_ORDER, $results_per_page = 0, $page = 0)
	{
		$value = $this->_db->qstr($value);
		if ($orderby == DBTABLE_USE_DEFAULT_ORDER) {
			$orderby = $this->_getOrderPreference();
		}
		if (empty($this->_fieldlist)) trigger_error('Empty fieldlist in table class',E_USER_ERROR);
		$sql = "SELECT $this->_fieldlist";
		if ($this->_key) $sql .= ",$this->_key";
		$sql .= ' FROM '.$this->_name." WHERE $field $clause '$value'";	//TODO: quote $field for Postgres?
		if ($orderby) $sql .= " ORDER BY $orderby";
		if ($results_per_page > 0) {
			return $this->_db->query_paged($sql, $results_per_page, $page);
		} else {
			return $this->_db->query($sql);
		}
	}
	
	// -----------------------------------------------------
	/**
	 * Delete all rows from the table where the contents of a specified column are as given.
	 *
	 * @param string $field
	 * @param string $value
	 * @return int  The number of affected rows
	 */
	function deleteRows ($field, $value)
	{
		$value = $this->_db->qstr($value);
		if (!$this->_name) trigger_error('DBTable::deleteRows() called on uninitialised table',E_USER_ERROR);
		if (!$field) trigger_error('No field name supplied for DBTable::deleteRows()',E_USER_ERROR);
		$res = $this->_db->query('DELETE FROM '.$this->_name." WHERE $field = '$value'");
		return $res->affectedRows();
	}
	
	// -----------------------------------------------------
	// Foreign keys
	
	function setForeignKeyDiscoveryType($Type)
	{
		$this->_ForeignKeyDiscoveryType = $Type;
	}
	
	function getForeignKeys()
	{
		// Send cached copy, if it exists
		if (isset($this->_ForeignKeyCache)) return $this->_ForeignKeyCache;
		
		$ForeignKeys = false;
		
		switch ($this->_ForeignKeyDiscoveryType) {
			case 'db':
				$ForeignKeys = _DBTable_getForeignKeys_db($this);
				break;
			case 'colnames':
				$ForeignKeys = $this->_getForeignKeysByColnames();
				break;
			case 'defined':
				$ForeignKeys = $this->ForeignKeys;
				break;
			default:
				trigger_error("DBTable::getForeignKeys() called, but foreign key discovery type is set to unknown type '$this->_ForeignKeyDiscoveryType'",E_USER_WARNING);
		}
		
		// Override/add to any discovered values by ones defined
		if ($this->_ForeignKeyDiscoveryType != 'defined' && isset($this->ForeignKeys)) {
			$ForeignKeys = $this->_getForeignKeyUnion($ForeignKeys);
		}
		
		// Save foreign key cache
		if ($ForeignKeys !== false) {
			$this->_ForeignKeyCache = $ForeignKeys;
		}
		return $ForeignKeys;		
	}
	
	function _getForeignKeyUnion($ForeignKeys)
	{
		if (is_array($this->ForeignKeys)) {
			foreach ($this->ForeignKeys as $Column => $Foreign) {
				$ForeignKeys[$Column] = $Foreign;
			}
		}
		return $ForeignKeys;
	}
	
	function _getForeignKeysByColnames()
	{
		$ForeignKeys = null;
		if (is_array($this->_fields)) {
			foreach ($this->_fields as $ColName) {
				if (preg_match("/^(.*)_id$/i",$ColName, $Matches)) {
					$ForeignKeys[$ColName]['table'] = $Matches[1].'s';
					$ForeignKeys[$ColName]['column'] = 'id';
				}
			}
		}
		return $ForeignKeys;
	}
	
	// -----------------------------------------------------
	// Misc. functions
	function getColumnDescription($ColName)
	{
		if (is_array($this->ColumnDescription) && isset($this->ColumnDescription[$ColName])) {
			$Description = $this->ColumnDescription[$ColName];
		} else {
			$Description = $ColName;
			$Description = preg_replace('/_id$/i','',$Description);
			$Description = str_replace('_',' ',$Description);
			$Description = ucfirst($Description);
		}
		return $Description;
	}
	
	// Useful when NOT using the old_value column in audit_actions
	// This will go through the table and do an initial insert into the audit log
	// for each row
	function insertInitialAuditRecords($owner=null)
	{
		$Row = new DBRow($this);
		$AllRows = $this->getAll();
		
		$this->_db->beginTransaction();
		while ($AllRows->fetchIntoObject($Row)) {
			
			// Clear the stored data for this row, so it looks like all 
			// columns have changed
			foreach ($this->_fields as $field) {
				// HACK: modifies priv property
				$Row->_audit[$field] = null;
			}
			
			if ($owner !== null) $Row->setAuditOwner($owner);
			$Row->setAuditNotes('Initial audit log creation');

			// HACK: calls priv method
			$Row->_auditLog(PHPOF_EVENT_INSERT);
		}
		$this->_db->commitTransaction();
	}
}


class DBColumn {
	var $Name = null;
	var $ForeignKey = false;
	var $Type = null;	// Unified PHPOF type
	var $Constraints = null;
	var $DefaultContent = null;
	var $IsPrimaryKey = false;
	
	function DBColumn ($Name=null)
	{
		if ($Name !== null) {
			$this->Name = $Name;
		}
	}
}
